
from scipy import *
from numpy import *
import numpy as np
from scipy.linalg import *
from pylab import *
from scipy.integrate import *

#This code solves for the motion of a chain with 
#1. fixed link lengths, 
#2. inertia is #negligible (it always is for microscopic systems in a solvent) 
#3 A drag on each monomer proportional to the velocity
#4 Various other forces can be added to the simulation, like giving it an elastic
#  bending rigity, or pulling on the ends.
#5 In another file, similar code is used to study electrophoresis by adding a periodic
# force field representing an array of obstacles.
# It can also be used to simulate the translocation of a chain through a pore.

#Because the links are constant in length, this makes it useful for studying many
#problems where the chain is close to being inextensible. However the algorithm
# to enforce inextensible links is a bit more involved. It involves calculating the
# tension along the chain as a function of position. Once the tension is calculated
# it is easy to add it to the other forces in the problem to get the equation of motion.

#This class is used to calculate the tension on the links of the chain.
class Tension:
    def __init__(self, T, n, d):
        self.T = T
        self.delta = zeros((n-1,d),dtype=double)
        self.delta_f = zeros((n-1,d),dtype=double)
        self.u = zeros((n-1),dtype=double)
        self.l = zeros((n-1),dtype=double)
        self.rhs = zeros((n-1,d),dtype=double)
        self.diag = zeros((n-1),dtype=double)
        self.force = zeros((n,d),dtype=double)


    def tension(self, r, f):
        self.delta = r[1:,:]-r[:-1,:]
        self.diag = -2*sum(self.delta*self.delta,axis=1)
        self.delta_f = f[1:,:]-f[:-1,:]
        u_vect = self.delta[1:,:]*self.delta[:-1,:]
        self.u[1:] = sum(u_vect, axis = 1)
        #self.u[1:] = sum(self.delta[1:,:]*self.delta[:-1,:],axis = 1)
        self.l[:-1] = self.u[1:]
        self.rhs = -sum(self.delta_f*self.delta,axis=1) - c1*(-0.5*self.diag - StepLengthSq)

        return solve_banded((1, 1), array((self.u,self.diag,self.l)), self.rhs, overwrite_ab=0, overwrite_b=0, debug=0)

    def calcForce(self,r,f):
        t = self.tension(r,f)
        for i in range(d):
            self.force[1:-1,i] = t[1:]*self.delta[1:,i] - t[0:-1]*self.delta[0:-1,i] + f[1:-1,i]
        self.force[0,:] = t[0]*self.delta[0,:] + f[0,:]
        self.force[n-1,:] = -t[n-2]*self.delta[n-2,:] + f[n-1,:]
        return self.force


def poreForce(r):
    distsq[:] = 0

    #get rid of overflow problem because first line below doesn't work:
    arg_[:] = (r[:,1]**2-y_0_sq)*invlambdasq
    arg_[:] = 700 + signbit(arg_[:]-700)*(arg_[:]-700)

    exp_y[:] = np.exp(arg_[:])
    exp_r[:] = np.exp((r_0_sq - r[:,0]**2 - r[:,2]**2)*invlambdasq)
    V_r[:] = 1.0/(1+exp_r[:])
    V_y[:] = 1.0/(1+exp_y[:])

    f_pore[:,0] = -V_0*r[:,0]*exp_r[:]*V_r[:]**2
    f_pore[:,2] = -V_0*r[:,2]*exp_r[:]*V_r[:]**2

    f_pore[:,1] = (V_0 * V_r[:] + 2*ampl*np.cos(k*r[:,1])) * r[:,1]*exp_y[:]*V_y[:]**2 + V_y[:]*ampl*k*np.sin(k*r[:,1]) + f_ext*V_y[:]
    return f_pore



ampl = 5.0
k = pi
f_ext = -10.0
def dx_dt(r_in,t):
    r = reshape(r_in,(n,d))
    f = noise + poreForce(r)
    return tens.calcForce(r,f).flatten()

def length_error(r):
    delta = r[1:,:] - r[:-1,:]
    dlt_sq = delta*delta
    return sum(sum(dlt_sq)) - (n-1)

def CenterOfMass(rout):
    rcm[:] = 0.0
    for i in range(n):
        rcm[:] += rout[i,:]
    return rcm/n


def construct_chain(n):
    r = zeros((1,d))
    r1 = zeros((1,d))
    shift = array([0.0,0.0,0.0])
    r_try = array([0,0,0.])
    for i in range(n-1):
        if i < 1.5*y_0:
            dlt = array([0.0,1.0,0.0])
            r_try = r1[0] + dlt
        else:
            while 1:
                dlt = standard_normal((d))
                dlt = StepLength*dlt/linalg.norm(dlt)
                r_try = r1[0] + dlt
                if r_try[1] > y_0:
                    break

        r1 = reshape(r_try,(1,d))
        
        r = append(r, r1, 0)
    r += shift
    return r

   

seed(3001)
StepLength = 1.0
StepLengthSq = StepLength**2
#dimension of space
d = 3
#number of beads on the chain
n = 128
dt = 0.05
#Temperature:
T = 2.0
c1 = 1 #helps to maintain constant link lengths
# parameters for the pore potential:
V_0 = 100.0
y_0 = 10.0
r_0 = 2.0
r_0_sq = r_0**2
y_0_sq = y_0**2
invlambdasq = (1.0/0.5)**2

a = 2.21
rcm = zeros((d))
f_pore = zeros((n,d))
distsq = zeros((n))
exp_r = zeros((n))
exp_y = zeros((n))
arg_ = zeros((n))
V_r = zeros((n))
V_y = zeros((n))



tens = Tension(T, n, d)

#r holds an array of the coordinates of the chain
r = construct_chain(n)

#time array
t = arange(0,2*dt,dt)

f = zeros((n,d))

graphics= True
if graphics:
    from vpython import *

    def vc(r):
        if d== 3:
           return vector(r[0],r[1],r[2])
        elif d==2:
           return vector(r[0],r[1],0)



    ball = []
    cyl = []
    for i in range(n-1):
        cyl.append(cylinder (pos=vc(r[i]), axis=vc(r[i+1]-r[i]), color=color.red, radius = 0.2))

import os
#if os.name != 'posix':#apparently we don't need to worry about the os name any more:
pore = cylinder(pos=vector(0,-y_0,0), axis = vector(0,2*y_0,0), color = color.blue, radius = 1.0*r_0, opacity = 0.5)
#else: #the lines below used to be necessary on the linux install, but now apparently it can handle transparency.
#pore_bottom = cylinder(pos=array([0,-y_0,0]), axis = array([0,0.5,0]), color = color.blue, radius = 1.0*r_0, opacity = 0.5)
#pore_top = cylinder(pos=array([0,y_0,0]), axis = array([0,-0.5,0]), color = color.blue, radius = 1.0*r_0, opacity = 0.5)



if graphics:
    ion()

x_out = r.flatten()

x_out = reshape(x_out,(1,n*d))

time = 0;
for t_coarse in arange(0,1000,dt):
    if graphics:
        rate(10000)
    t = array([t_coarse,t_coarse+dt])
    #thermal noise is generated and used to push the chain:
    noise = sqrt(2*T/dt)*standard_normal((n,d))
    #The diff'lequations are integrated. All the work comes in calculating dx_dt
    x_out = odeint(dx_dt, x_out[-1] , t, full_output=False)
    time += 1
    if time%300 == 0 : 
        #check to make sure the links haven't changed their lengths
        print(length_error(reshape(x_out[-1],(n,d))))
    if  graphics:
        r_out = reshape(x_out[-1],(n,d))
        for i in range(n-1):
            cyl[i].pos = vc(r_out[i])
            cyl[i].axis = vc(r_out[i+1]-r_out[i])
        print("t = ", t_coarse, "Center of Mass = ", CenterOfMass(r_out))



